@@ -21,12 +21,13 @@ from tqdm.auto import tqdm
 
 from .argument_utility import (
     ActionScalerArg,
+    RewardScalerArg,
     ScalerArg,
     UseGPUArg,
     check_action_scaler,
+    check_reward_scaler,
     check_scaler,
 )
-from .augmentation import AugmentationPipeline, DrQPipeline, create_augmentation
 from .constants import (
     CONTINUOUS_ACTION_SPACE_MISMATCH_ERROR,
     DISCRETE_ACTION_SPACE_MISMATCH_ERROR,
@@ -46,8 +47,10 @@ from .models.q_functions import QFunctionFactory, create_q_func_factory
 from .online.utility import get_action_size_from_env
 from .preprocessing import (
     ActionScaler,
+    RewardScaler,
     Scaler,
     create_action_scaler,
+    create_reward_scaler,
     create_scaler,
 )
 
@@ -77,7 +80,14 @@ def _serialize_params(params: Dict[str, Any]) -> Dict[str, Any]:
         if isinstance(value, Device):
             params[key] = value.get_id()
         elif isinstance(
-            value, (Scaler, ActionScaler, EncoderFactory, QFunctionFactory)
+            value,
+            (
+                Scaler,
+                ActionScaler,
+                RewardScaler,
+                EncoderFactory,
+                QFunctionFactory,
+            ),
         ):
             params[key] = {
                 "type": value.get_type(),
@@ -85,14 +95,6 @@ def _serialize_params(params: Dict[str, Any]) -> Dict[str, Any]:
             }
         elif isinstance(value, OptimizerFactory):
             params[key] = value.get_params()
-        elif isinstance(value, AugmentationPipeline):
-            aug_types = value.get_augmentation_types()
-            aug_params = value.get_augmentation_params()
-            params[key] = {"params": value.get_params(), "augmentations": []}
-            for aug_type, aug_param in zip(aug_types, aug_params):
-                params[key]["augmentations"].append(
-                    {"type": aug_type, "params": aug_param}
-                )
     return params
 
 
@@ -103,19 +105,16 @@ def _deseriealize_params(params: Dict[str, Any]) -> Dict[str, Any]:
             scaler_params = params["scaler"]["params"]
             scaler = create_scaler(scaler_type, **scaler_params)
             params[key] = scaler
-        if key == "action_scaler" and params["action_scaler"]:
+        elif key == "action_scaler" and params["action_scaler"]:
             scaler_type = params["action_scaler"]["type"]
             scaler_params = params["action_scaler"]["params"]
             action_scaler = create_action_scaler(scaler_type, **scaler_params)
             params[key] = action_scaler
-        elif key == "augmentation" and params["augmentation"]:
-            augmentations = []
-            for param in params[key]["augmentations"]:
-                aug_type = param["type"]
-                aug_params = param["params"]
-                augmentation = create_augmentation(aug_type, **aug_params)
-                augmentations.append(augmentation)
-            params[key] = DrQPipeline(augmentations, **params[key]["params"])
+        elif key == "reward_scaler" and params["reward_scaler"]:
+            scaler_type = params["reward_scaler"]["type"]
+            scaler_params = params["reward_scaler"]["params"]
+            reward_scaler = create_reward_scaler(scaler_type, **scaler_params)
+            params[key] = reward_scaler
         elif "optim_factory" in key:
             params[key] = OptimizerFactory(**value)
         elif "encoder_factory" in key:
@@ -138,12 +137,14 @@ class LearnableBase:
     _gamma: float
     _scaler: Optional[Scaler]
     _action_scaler: Optional[ActionScaler]
+    _reward_scaler: Optional[RewardScaler]
     _real_ratio: float
     _generated_maxlen: int
     _impl: Optional[ImplBase]
     _eval_results: DefaultDict[str, List[float]]
     _loss_history: DefaultDict[str, List[float]]
     _active_logger: Optional[D3RLPyLogger]
+    _grad_step: int
 
     def __init__(
         self,
@@ -151,8 +152,9 @@ class LearnableBase:
         n_frames: int,
         n_steps: int,
         gamma: float,
-        scaler: ScalerArg,
-        action_scaler: ActionScalerArg,
+        scaler: ScalerArg = None,
+        action_scaler: ActionScalerArg = None,
+        reward_scaler: RewardScalerArg = None,
         real_ratio: float = 1.0,
         generated_maxlen: int = 100000,
         kwargs: Optional[Dict[str, Any]] = None,
@@ -163,6 +165,7 @@ class LearnableBase:
         self._gamma = gamma
         self._scaler = check_scaler(scaler)
         self._action_scaler = check_action_scaler(action_scaler)
+        self._reward_scaler = check_reward_scaler(reward_scaler)
         self._real_ratio = real_ratio
         self._generated_maxlen = generated_maxlen
 
@@ -170,6 +173,7 @@ class LearnableBase:
         self._eval_results = defaultdict(list)
         self._loss_history = defaultdict(list)
         self._active_logger = None
+        self._grad_step = 0
 
         if kwargs and len(kwargs.keys()) > 0:
             LOG.warning("Unused arguments are passed.", **kwargs)
@@ -290,6 +294,9 @@ class LearnableBase:
                 "_eval_results",
                 "_loss_history",
                 "_active_logger",
+                "_grad_step",
+                "active_logger",
+                "grad_step",
                 "observation_shape",
                 "action_size",
             ]:
@@ -358,6 +365,7 @@ class LearnableBase:
             Dict[str, Callable[[Any, List[Episode]], float]]
         ] = None,
         shuffle: bool = True,
+        callback: Optional[Callable[["LearnableBase", int, int], None]] = None,
     ) -> List[Tuple[int, Dict[str, float]]]:
         """Trains with the given dataset.
 
@@ -388,6 +396,8 @@ class LearnableBase:
             save_interval: interval to save parameters.
             scorers: list of scorer functions used with `eval_episodes`.
             shuffle: flag to shuffle transitions on each epoch.
+            callback: callable function that takes ``(algo, epoch, total_step)``
+                , which is called every step.
 
         Returns:
             list of result tuples (epoch, metrics) per epoch.
@@ -410,6 +420,7 @@ class LearnableBase:
                 save_interval,
                 scorers,
                 shuffle,
+                callback,
             )
         )
         return results
@@ -433,6 +444,7 @@ class LearnableBase:
             Dict[str, Callable[[Any, List[Episode]], float]]
         ] = None,
         shuffle: bool = True,
+        callback: Optional[Callable[["LearnableBase", int, int], None]] = None,
     ) -> Generator[Tuple[int, Dict[str, float]], None, None]:
         """Iterate over epochs steps to train with the given dataset. At each
              iteration algo methods and properties can be changed or queried.
@@ -466,6 +478,8 @@ class LearnableBase:
             save_interval: interval to save parameters.
             scorers: list of scorer functions used with `eval_episodes`.
             shuffle: flag to shuffle transitions on each epoch.
+            callback: callable function that takes ``(algo, epoch, total_step)``
+                , which is called every step.
 
         Returns:
             iterator yielding current epoch and metrics dict.
@@ -491,7 +505,7 @@ class LearnableBase:
 
         iterator: TransitionIterator
         if n_epochs is None and n_steps is not None:
-            assert n_steps > n_steps_per_epoch
+            assert n_steps >= n_steps_per_epoch
             n_epochs = n_steps // n_steps_per_epoch
             iterator = RandomIterator(
                 episodes,
@@ -545,16 +559,26 @@ class LearnableBase:
             )
             self._action_scaler.fit(episodes)
 
+        # initialize reward scaler
+        if self._reward_scaler:
+            LOG.debug(
+                "Fitting reward scaler...",
+                reward_scaler=self._reward_scaler.get_type(),
+            )
+            self._reward_scaler.fit(episodes)
+
         # instantiate implementation
         if self._impl is None:
-            LOG.debug("Building model...")
+            LOG.debug("Building models...")
             transition = iterator.transitions[0]
             action_size = transition.get_action_size()
             observation_shape = tuple(transition.get_observation_shape())
             self.create_impl(
                 self._process_observation_shape(observation_shape), action_size
             )
-            LOG.debug("Model has been built.")
+            LOG.debug("Models have been built.")
+        else:
+            LOG.warning("Skip building models since they're already built.")
 
         # save hyperparameters
         self.save_params(logger)
@@ -584,14 +608,14 @@ class LearnableBase:
 
                 # generate new transitions with dynamics models
                 new_transitions = self.generate_new_data(
-                    epoch=epoch,
-                    total_step=total_step,
                     transitions=iterator.transitions,
                 )
                 if new_transitions:
                     iterator.add_generated_transitions(new_transitions)
                     LOG.debug(
-                        f"{len(new_transitions)} transitions are generated."
+                        f"{len(new_transitions)} transitions are generated.",
+                        real_transitions=len(iterator.transitions),
+                        fake_transitions=len(iterator.generated_transitions),
                     )
 
                 with logger.measure_time("step"):
@@ -601,7 +625,7 @@ class LearnableBase:
 
                     # update parameters
                     with logger.measure_time("algorithm_update"):
-                        loss = self.update(epoch, total_step, batch)
+                        loss = self.update(batch)
 
                     # record metrics
                     for name, val in loss.items():
@@ -615,7 +639,11 @@ class LearnableBase:
                         }
                         range_gen.set_postfix(mean_loss)
 
-                    total_step += 1
+                total_step += 1
+
+                # call callback if given
+                if callback:
+                    callback(self, epoch, total_step)
 
             # save loss to loss history dict
             self._loss_history["epoch"].append(epoch)
@@ -632,7 +660,7 @@ class LearnableBase:
 
             # save model parameters
             if epoch % save_interval == 0:
-                logger.save_model(epoch, self)
+                logger.save_model(total_step, self)
 
             yield epoch, metrics
 
@@ -697,32 +725,31 @@ class LearnableBase:
             observation_shape = (self._n_frames * n_channels, *image_size)
         return observation_shape
 
-    def update(
-        self, epoch: int, total_step: int, batch: TransitionMiniBatch
-    ) -> Dict[str, float]:
+    def update(self, batch: TransitionMiniBatch) -> Dict[str, float]:
         """Update parameters with mini-batch of data.
 
         Args:
-            epoch: the current number of epochs.
-            total_step: the current number of total iterations.
             batch: mini-batch data.
 
         Returns:
             dictionary of metrics.
 
         """
+        loss = self._update(batch)
+        self._grad_step += 1
+        return loss
+
+    def _update(self, batch: TransitionMiniBatch) -> Dict[str, float]:
         raise NotImplementedError
 
     def generate_new_data(
-        self, epoch: int, total_step: int, transitions: List[Transition]
+        self, transitions: List[Transition]
     ) -> Optional[List[Transition]]:
         """Returns generated transitions for data augmentation.
 
         This method is for model-based RL algorithms.
 
         Args:
-            epoch: the current epoch.
-            total_step: the total update steps.
             transitions: list of transitions.
 
         Returns:
@@ -900,6 +927,20 @@ class LearnableBase:
     def action_scaler(self, action_scaler: ActionScaler) -> None:
         self._action_scaler = action_scaler
 
+    @property
+    def reward_scaler(self) -> Optional[RewardScaler]:
+        """Preprocessing reward scaler.
+
+        Returns:
+            Optional[RewardScaler]: preprocessing reward scaler.
+
+        """
+        return self._reward_scaler
+
+    @reward_scaler.setter
+    def reward_scaler(self, reward_scaler: RewardScaler) -> None:
+        self._reward_scaler = reward_scaler
+
     @property
     def impl(self) -> Optional[ImplBase]:
         """Implementation object.
@@ -937,3 +978,50 @@ class LearnableBase:
         if self._impl:
             return self._impl.action_size
         return None
+
+    @property
+    def active_logger(self) -> Optional[D3RLPyLogger]:
+        """Active D3RLPyLogger object.
+
+        This will be only available during training.
+
+        Returns:
+            logger object.
+
+        """
+        return self._active_logger
+
+    def set_active_logger(self, logger: D3RLPyLogger) -> None:
+        """Set active D3RLPyLogger object
+
+        Args:
+            logger: logger object.
+
+        """
+        self._active_logger = logger
+
+    @property
+    def grad_step(self) -> int:
+        """Total gradient step counter.
+
+        This value will keep counting after ``fit`` and ``fit_online``
+        methods finish.
+
+        Returns:
+            total gradient step counter.
+
+        """
+        return self._grad_step
+
+    def set_grad_step(self, grad_step: int) -> None:
+        """Set total gradient step counter.
+
+        This method can be used to restart the middle of training with an
+        arbitrary gradient step counter, which has effects on periodic
+        functions such as the target update.
+
+        Args:
+            grad_step: total gradient step counter.
+
+        """
+        self._grad_step = grad_step
