@@ -16,9 +16,7 @@ from .explorers import Explorer
 
 
 class AlgoProtocol(Protocol):
-    def update(
-        self, epoch: int, total_step: int, batch: TransitionMiniBatch
-    ) -> Dict[str, float]:
+    def update(self, batch: TransitionMiniBatch) -> Dict[str, float]:
         ...
 
     def build_with_env(self, env: gym.Env) -> None:
@@ -36,6 +34,9 @@ class AlgoProtocol(Protocol):
     def save_model(self, fname: str) -> None:
         ...
 
+    def set_active_logger(self, logger: D3RLPyLogger) -> None:
+        ...
+
     @property
     def action_size(self) -> Optional[int]:
         ...
@@ -68,6 +69,10 @@ class AlgoProtocol(Protocol):
     def impl(self) -> Optional[Any]:
         ...
 
+    @property
+    def grad_step(self) -> int:
+        ...
+
 
 def _setup_algo(algo: AlgoProtocol, env: gym.Env) -> None:
     # initialize scaler
@@ -88,6 +93,8 @@ def _setup_algo(algo: AlgoProtocol, env: gym.Env) -> None:
         LOG.debug("Building model...")
         algo.build_with_env(env)
         LOG.debug("Model has been built.")
+    else:
+        LOG.warning("Skip building models since they're already built.")
 
 
 def train_single_env(
@@ -157,6 +164,7 @@ def train_single_env(
         tensorboard_dir=tensorboard_dir,
         with_timestamp=with_timestamp,
     )
+    algo.set_active_logger(logger)
 
     # initialize algorithm parameters
     _setup_algo(algo, env)
@@ -245,16 +253,16 @@ def train_single_env(
 
                     # update parameters
                     with logger.measure_time("algorithm_update"):
-                        loss = algo.update(
-                            epoch=epoch,
-                            total_step=total_step // update_interval,
-                            batch=batch,
-                        )
+                        loss = algo.update(batch)
 
                     # record metrics
                     for name, val in loss.items():
                         logger.add_metric(name, val)
 
+            # call callback if given
+            if callback:
+                callback(algo, epoch, total_step)
+
         if epoch > 0 and total_step % n_steps_per_epoch == 0:
             # evaluation
             if eval_scorer:
@@ -263,10 +271,6 @@ def train_single_env(
             if epoch % save_interval == 0:
                 logger.save_model(total_step, algo)
 
-            # call callback if given
-            if callback:
-                callback(algo, epoch, total_step)
-
             # save metrics
             logger.commit(epoch, total_step)
 
@@ -338,6 +342,7 @@ def train_batch_env(
         tensorboard_dir=tensorboard_dir,
         with_timestamp=with_timestamp,
     )
+    algo.set_active_logger(logger)
 
     # initialize algorithm parameters
     _setup_algo(algo, env)
@@ -412,6 +417,10 @@ def train_batch_env(
                 if clip_episode[i] and is_image:
                     stacked_frame.clear_by_index(i)
 
+            # call callback if given
+            if callback:
+                callback(algo, epoch, total_step)
+
         for step in range(n_updates_per_epoch):
             # sample mini-batch
             with logger.measure_time("sample_batch"):
@@ -424,11 +433,7 @@ def train_batch_env(
 
             # update parameters
             with logger.measure_time("algorithm_update"):
-                loss = algo.update(
-                    epoch=epoch,
-                    total_step=epoch * n_updates_per_epoch + step,
-                    batch=batch,
-                )
+                loss = algo.update(batch)
 
             # record metrics
             for name, val in loss.items():
@@ -439,10 +444,6 @@ def train_batch_env(
             if eval_scorer:
                 logger.add_metric("evaluation", eval_scorer(algo))
 
-        # call callback if given
-        if callback:
-            callback(algo, epoch, total_step)
-
         # save metrics
         logger.commit(epoch, total_step)
 
